
// Subscripts: ranges and embeddings

#include "RangeRep.h"
#include "StrideRep.h"
#include "CollapsedRangeRep.h"
#include "BlockRangeRep.h"
#include "CyclicRangeRep.h"
#include "BlockCyclicRangeRep.h"
#include "StepRangeRep.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Range': handle class for range representations.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

enum Format {BLK, CYC, STP} ;

class Stride ;

class Range {
public :
  Range() : rep(0) {}

  Range(const Range& x) : rep(x.rep) {
    if(rep)
      rep->count++ ;
  }

  Range& operator=(const Range& x) {
    if(x.rep) 
      x.rep->count++ ;
    if(rep && --rep->count == 0)
      delete rep ;
    rep = x.rep ;
    return *this ;
  }

  Range(const int extent, const Layout layout = STANDARD) {
    rep = new CollapsedRangeRep(extent, layout) ;
    rep->count = 1 ;
  }

  Range(Dimension dim) {
    rep = new BlockRangeRep(dim.size(), dim, 1) ;
        // Should add `DimRangeRep' if this proves useful.
    rep->count = 1 ;
  }

  Range(const int extent, Dimension dim, const Format fmt,
        const Layout layout = STANDARD) {
    switch(fmt) {
      case BLK : {
        const int blockSize = (extent + dim.size() - 1) / dim.size() ;
        if(layout == STANDARD)
          rep = new BlockRangeRep(extent, dim, blockSize) ;
        else
          rep = new PackedBlockRangeRep(extent, dim, blockSize) ;

        break ;
      }
      case CYC :
        rep = new CyclicRangeRep(extent, dim, layout) ;
        break ;
      case STP :
        rep = new StepRangeRep(extent, dim, layout) ;
        break ;
    }
    rep->count = 1 ;
  }

  Range(const int extent, Dimension dim, const Format fmt, const int blockSize,
        const Layout layout = STANDARD) {
    switch(fmt) {
      case BLK :
        if(extent > dim.size() * blockSize)
            AdlibError("Range :: Range: "
                       "extent too large for specified block size") ;
        if(layout == STANDARD)
          rep = new BlockRangeRep(extent, dim, blockSize) ;
        else
          rep = new PackedBlockRangeRep(extent, dim, blockSize) ;
        break ;
      case CYC :
        rep = new BlockCyclicRangeRep(extent, dim, blockSize, layout) ;
        break ;
    }
    rep->count = 1 ;
  }

  Range(const int extent, Dimension dim, const Format fmt,
        const Layout layout, const int wlo, const int whi) {
    switch(fmt) {
      case BLK :
	switch(layout) {
	  case SHADOW :
	    const int blockSize = (extent + dim.size() - 1) / dim.size() ;
	    rep = new ExtendedBlockRangeRep(extent, dim, blockSize, wlo, whi) ;
	    break ;
        }
        break ;
    }
    rep->count = 1 ;
  }

  ~Range() {
    if(rep && --rep->count == 0)
      delete rep ;
  }

  // Inquiries

  int size() const {return rep->size() ; }

  Dimension dim() const {return rep->dim ; }

  Range prm() const {
    Range result ;

    result.rep = rep->prm() ;
    result.rep->count++ ;

    return result ;
  }

  int bas() const {return rep->bas() ; }

  int str() const {return rep->str() ; }

  // Subscripting

  Subscript& operator()(const int i) {
    return *rep->subscr(i) ;
  }

  int operator()(Subscript& s) {
    return rep->subscr(&s) ;
  }

  Range subrng(const int extent, const int base, const int stride = 1) const {
    if(extent != 0) {
      const int parExtent = size() ;
      const int top = base + stride * (extent - 1) ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: subrng: subrange out of range") ;
    }

    Range result ;
    result.rep = rep->subrng(extent, base, stride) ;
    result.rep->count = 1 ;
    return result ;
  }

  int step(Block& i) const ;

  // `Stride' ``constructors''

  Stride vector(const int dataLen) const ;

  Stride promote(const int oldStr) const ;

  // Lower level iterations...

  void blockInit(Block& b) {rep->blockInit(b) ; }
  void locBlockInit(Block& b) {rep->locBlockInit(b) ; }

  void locBlocksEnum(Block& b) {rep->locBlocksEnum(b) ; }
  int  locBlocksTest(Block& b) {return rep->locBlocksTest(b) ; }
  void locBlocksNext(Block& b) {rep->locBlocksNext(b) ; }

  void allBlocksEnum(Block& b) {rep->allBlocksEnum(b) ; }
  int  allBlocksTest(Block& b) {return rep->allBlocksTest(b) ; }
  void allBlocksNext(Block& b) {rep->allBlocksNext(b) ; }

  // Interim support for PCRC Fortran interface

  int& g_lb() {return rep->g_lb ; }
  int& g_stride() {return rep->g_stride ; }
private :
  RangeRep* rep ;
} ;


// Iterator (index) classes

struct Block : public Subscript {

  Block(const Range& _parent) : parent(_parent) {parent.blockInit(*this) ; }

  Range parent ;

  int blockCount ;

  int blockBase ;
  int blockStep ;

  int blockBaseSub ;
  int blockStepSub ;
} ;

struct LocBlocksIndex : public Block {
  LocBlocksIndex(const Range& _parent) : Block(_parent) {
    parent.locBlockInit(*this) ;
    sub = blockBaseSub ;
  }

  void begin() {
    parent.locBlocksEnum(*this) ;
    sub = blockBaseSub ;
  }
  int  test() {return parent.locBlocksTest(*this) ; }
  void next() {
    parent.locBlocksNext(*this) ;
    sub = blockBaseSub ;
  }
} ;

struct AllBlocksIndex : public Block {
  AllBlocksIndex(const Range& _parent) : Block(_parent) {}

  void begin() {
    parent.allBlocksEnum(*this) ;
    sub = blockBaseSub ;
  }
  int  test() {return parent.allBlocksTest(*this) ; }
  void next() {
    parent.allBlocksNext(*this) ;
    sub = blockBaseSub ;
  }
} ;

struct Index : public LocBlocksIndex {
public :
  Index(Range _parent) : LocBlocksIndex(_parent) {}

  // Subscripting

  operator int() const {return i ; }

  // `where' iterators

  void where() {
    //at() ;
    if(dim) dim.where() ;
    LocBlocksIndex :: begin() ;
    l   = 0 ;
    i   = blockBase ;
    //sub = blockBaseSub ;
  }

  int test() {
    return LocBlocksIndex :: test() ;
  }

  void next() {
    l++ ;
    if(l == blockCount) {
      LocBlocksIndex :: next() ;
      l   = 0 ;
      i   = blockBase ;
      //sub = blockBaseSub ;
    }
    else {
      i   += blockStep ;
      sub += blockStepSub ;
    }
  }

  void erewh() {
    //ta() ;
    if(dim) dim.erewh() ;
  }

private :
  int i, l ;

public :

  // Cosmetic support for ghost regions:

  Subscript& operator<<(const int shift) {
    shifted = *this ;
    shifted.sub -= blockStepSub * shift ;
    return shifted ;
  }

  Subscript& operator>>(const int shift) {
    shifted = *this ;
    shifted.sub += blockStepSub * shift ;
    return shifted ;
  }

private :

  Subscript shifted ;
} ;

// GNU compiler needs these

inline int operator+(const Index& i, const int n) {
  return (int) i + n ;
}

inline int operator+(const int n, const Index& i) {
  return n + (int) i ;
}

inline int operator+(const Index& i, const Index& j) {
  return (int) i + (int) j ;
}

inline int operator-(const Index& i, const int n) {
  return (int) i - n ;
}

inline int operator-(const int n, const Index& i) {
  return n - (int) i ;
}

inline int operator-(const Index& i, const Index& j) {
  return (int) i - (int) j ;
}

inline int operator*(const Index& i, const int n) {
  return (int) i * n ;
}

inline int operator*(const int n, const Index& i) {
  return n * (int) i ;
}

inline int operator*(const Index& i, const Index& j) {
  return (int) i * (int) j ;
}

inline int operator/(const Index& i, const int n) {
  return (int) i / n ;
}

inline int operator/(const int n, const Index& i) {
  return n / (int) i ;
}

inline int operator/(const Index& i, const Index& j) {
  return (int) i / (int) j ;
}




inline int RangeRep :: step(Block& i) {
  return i.blockStep * i.parent.str() / str() ;
}

inline int Range :: step(Block& i) const {
  return rep->step(i) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Stride': Handle class for `StrideRep'.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Stride {
public :
  Stride() : rep(0) {}

  Stride(const Stride& str) : rep(str.rep) {
    if(rep)
      rep->count++ ;
  }

  Stride& operator=(const Stride& str) {
    if(str.rep) 
      str.rep->count++ ;
    if(rep && --rep->count == 0)
      delete rep ;
    rep = str.rep ;
    return *this ;
  }

  ~Stride() {
    if(rep && --rep->count == 0)
      delete rep ;
  }

  int offset(const Subscript& subs) const {
    return rep ? rep->offset(&subs) : 0 ;
  }

  int step(Block& i) const {
    return rep ? rep->step(i) : 0 ;
  }

  int volume() const {
    return rep ? rep->volume() : 0 ;
  }

  operator int() const {return rep->memStep ; }

private :
  StrideRep* rep ;
friend class Range ;
} ;

inline Stride Range :: vector(const int dataLen) const {
  Stride str ;
  str.rep = rep->vector(dataLen) ;
  str.rep->count = 1 ;
  return str ;
}

inline Stride Range :: promote(const int oldStr) const {
  Stride str ;
  str.rep = new PromoteStrideRep(rep, oldStr) ;
  str.rep->count = 1 ;
  return str ;
}



// ---------------------------------------------------------------
//
// Things to do:
//
//   Change signature of `Stride :: offset(const Subscript& subs)' to
//   `Stride :: operator()(Subscript& s)'???
//
//   Revert to setting `sub' directly in iterators?  It's a pain to have
//          i.sub = i.blockBaseSub ;
//   at the start of almost every loop body.
//   Maybe just identify `blockBaseSub' with `sub' after all?
//
//   Add a new `RangeRep' which allows a `Dimension' to be treated as a
//   range.
//
//   Change `subrng' to take lower-bound/upper-bound type parametrization???

